"""
These elements are used by the translator to construct the model from the
interpreted results. It is technically possible to use these functions to
build a model from scratch. But - it would be rather error prone.

This is code to assemble a pysd model once all of the elements have
been translated from their native language into python compatible syntax.
There should be nothing here that has to know about either vensim or
xmile specific syntax.
"""

import os.path
import textwrap
import warnings
from io import open
import black
import json

from . import utils

from pysd._version import __version__


class Imports():
    """
    Class to save the imported modules information for intelligent import
    """
    _numpy, _xarray, _subs = False, False, False
    _functions, _external, _utils = set(), set(), set()

    @classmethod
    def add(cls, module, function=None):
        """
        Add a function from module.

        Parameters
        ----------
        module: str
          module name.

        function: str or None
          function name. If None module will be set to true.

        """
        if function:
            getattr(cls, f"_{module}").add(function)
        else:
            setattr(cls, f"_{module}", True)

    @classmethod
    def get_header(cls, outfile, force_root=False):
        """
        Returns the importing information to print in the model file

        Parameters
        ----------
        outfile: str
            Name of the outfile to print in the header.

        force_root: bool (optional)
            If True, the _root variable will be returned to include in the
            model file and os.path will be imported. If False, the _root
            variable will only be included if the model has External
            objects.

        Returns
        -------
        text: str
            Header of the translated model file.

        """
        text =\
            f'"""\nPython model \'{outfile}\'\nTranslated using PySD\n"""\n\n'

        _root = ""

        if cls._external or force_root:
            # define root only if needed
            text += "from os import path\n"
            _root = "\n    _root = path.dirname(__file__)\n"
        if cls._numpy:
            text += "import numpy as np\n"
        if cls._xarray:
            text += "import xarray as xr\n"
        text += "\n"

        if cls._functions:
            text += "from pysd.py_backend.functions import %(methods)s\n"\
                    % {"methods": ", ".join(cls._functions)}
        if cls._external:
            text += "from pysd.py_backend.external import %(methods)s\n"\
                    % {"methods": ", ".join(cls._external)}
        if cls._utils:
            text += "from pysd.py_backend.utils import %(methods)s\n"\
                    % {"methods": ", ".join(cls._utils)}

        if cls._subs:
            text += "from pysd import cache, subs\n"
        else:
            # we need to import always cache as it is called in the integration
            text += "from pysd import cache\n"

        cls.reset()

        return text, _root

    @classmethod
    def reset(cls):
        """
        Reset the imported modules
        """
        cls._numpy, cls._xarray, cls._subs = False, False, False
        cls._functions, cls._external, cls._utils = set(), set(), set()


# Variable to save identifiers of external objects
build_names = set()


def build_modular_model(elements, subscript_dict, namespace, main_filename,
                        elements_per_view):

    """
    This is equivalent to the build function, but is used when the
    split_views parameter is set to True in the read_vensim function.
    The main python model file will be named as the original model file,
    and stored in the same folder. The modules will be stored in a separate
    folder named modules + original_model_name. Three extra json files will
    be generated, containing the namespace, subscripts_dict and the module
    names plus the variables included in each module, respectively.

    Setting split_views=True is recommended for large models with many
    different views.

    Parameters
    ----------
    elements: list
        Each element is a dictionary, with the various components needed
        to assemble a model component in python syntax. This will contain
        multiple entries for elements that have multiple definitions in
        the original file, and which need to be combined.

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    namespace: dict
        Translation from original model element names (keys) to python safe
        function identifiers (values).

    main_filename: str
        The name of the file to write the main module of the model to.

    elements_per_module: dict
        Contains the names of the modules as keys and the variables in
        each specific module inside a list as values.

    """
    root_dir = os.path.dirname(main_filename)
    model_name = os.path.basename(main_filename).split(".")[0]
    modules_dir = os.path.join(root_dir, "modules_" + model_name)
    # create modules directory if it does not exist
    os.makedirs(modules_dir, exist_ok=True)

    # check if there are subviews or only main views
    subviews = all(isinstance(n, dict) for n in elements_per_view.values())

    all_views = elements_per_view.keys()
    # creating the rest of files per module (this needs to be run before the
    # main module, as it updates the import_modules)
    processed_elements = []
    for view_name in all_views:
        view_elems = []
        if not subviews:  # only main views
            for element in elements:
                if element.get("py_name", None) in \
                   elements_per_view[view_name] or \
                   element.get("parent_name", None) in \
                   elements_per_view[view_name]:
                    view_elems.append(element)

            _build_separate_module(view_elems, subscript_dict, view_name,
                                   modules_dir)

        else:
            # create subdirectory
            view_dir = os.path.join(modules_dir, view_name)
            os.makedirs(view_dir, exist_ok=True)

            for subview_name in elements_per_view[view_name].keys():
                subview_elems = []
                for element in elements:
                    if element.get("py_name", None) in \
                       elements_per_view[view_name][subview_name] or \
                       element.get("parent_name", None) in \
                       elements_per_view[view_name][subview_name]:
                        subview_elems.append(element)

                _build_separate_module(subview_elems, subscript_dict,
                                       subview_name, view_dir)
                view_elems += subview_elems

        processed_elements += view_elems

    # the unprocessed will go in the main file
    unprocessed_elements = [
        element for element in elements if element not in processed_elements
    ]
    # building main file using the build function
    _build_main_module(unprocessed_elements, subscript_dict,
                       main_filename, subviews)

    # create json file for the modules and corresponding model elements
    with open(os.path.join(modules_dir, "_modules.json"), "w") as outfile:
        json.dump(elements_per_view, outfile, indent=4, sort_keys=True)

    # create single namespace in a separate json file
    with open(
        os.path.join(root_dir, "_namespace_" + model_name + ".json"), "w"
    ) as outfile:
        json.dump(namespace, outfile, indent=4, sort_keys=True)

    # create single subscript_dict in a separate json file
    with open(
        os.path.join(root_dir, "_subscripts_" + model_name + ".json"), "w"
    ) as outfile:
        json.dump(subscript_dict, outfile, indent=4, sort_keys=True)


def _build_main_module(elements, subscript_dict, file_name, subviews):
    """
    Constructs and writes the python representation of the main model
    module, when the split_views=True in the read_vensim function.

    Parameters
    ----------
    elements: list
        Elements belonging to the main module. Ideally, there should only be
        the initial_time, final_time, saveper and time_step, functions, though
        there might be others in some situations. Each element is a
        dictionary, with the various components needed to assemble a model
        component in python syntax. This will contain multiple entries for
        elements that have multiple definitions in the original file, and
        which need to be combined.

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    file_name: str
        Path of the file where the main module will be stored.

    subviews: bool
        True or false depending on whether the views are split in subviews or
        not.

    Returns
    -------
    None or text: None or str
        If file_name="return" it will return the content of the output file
        instead of saving it. It is used for testing.

    """
    all_elements = merge_partial_elements(elements)
    # separating between control variables and rest of variables
    control_vars_ = [element for element in all_elements if
                     element["py_name"] in ["initial_time",
                                            "final_time",
                                            "time_step",
                                            "saveper"]]
    elements = [element for element in all_elements if element not in
                control_vars_]

    control_vars = _generate_functions(control_vars_, subscript_dict)
    funcs = _generate_functions(elements, subscript_dict)

    Imports.add("utils", "load_model_data")
    Imports.add("utils", "open_module")

    # import of needed functions and packages
    text, root = Imports.get_header(os.path.basename(file_name),
                                    force_root=True)

    # import namespace from json file
    text += textwrap.dedent("""
    __pysd_version__ = '%(version)s'

    __data = {
        'scope': None,
        'time': lambda: 0
    }
    %(root)s
    _namespace, _subscript_dict, _modules = load_model_data(_root,
    "%(outfile)s")
    """ % {
        "outfile": os.path.basename(file_name).split(".")[0],
        "root": root,
        "version": __version__
    })

    text += _get_control_vars(control_vars)

    if not subviews:
        text += textwrap.dedent("""
        # load modules from the modules_%(outfile)s directory
        for module in _modules:
            exec(open_module(_root, "%(outfile)s", module))

        """ % {
            "outfile": os.path.basename(file_name).split(".")[0],
        })
    else:
        text += textwrap.dedent("""
        # load submodules from subdirs in modules_%(outfile)s directory
        for mod_name, mod_submods in _modules.items():
            for submod_name in mod_submods.keys():
                exec(open_module(_root, "%(outfile)s", mod_name, submod_name))

        """ % {
            "outfile": os.path.basename(file_name).split(".")[0],
        })

    text += funcs
    text = black.format_file_contents(text, fast=True, mode=black.FileMode())

    # Needed for various sessions
    build_names.clear()

    with open(file_name, "w", encoding="UTF-8") as out:
        out.write(text)


def _build_separate_module(elements, subscript_dict, module_name, module_dir):
    """
    Constructs and writes the python representation of a specific model
    module, when the split_views=True in the read_vensim function

    Parameters
    ----------
    elements: list
        Elements belonging to the module module_name. Each element is a
        dictionary, with the various components needed to assemble a model
        component in python syntax. This will contain multiple entries for
        elements that have multiple definitions in the original file, and
        which need to be combined.

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    module_name: str
        Name of the module

    module_dir: str
        Path of the directory where module files will be stored.

    Returns
    -------
    None

    """
    text = textwrap.dedent('''
    """
    Module %(module_name)s
    Translated using PySD version %(version)s
    """
    ''' % {
        "module_name": module_name,
        "version": __version__,
    })
    elements = merge_partial_elements(elements)
    funcs = _generate_functions(elements, subscript_dict)
    text += funcs
    text = black.format_file_contents(text, fast=True, mode=black.FileMode())

    outfile_name = os.path.join(module_dir, module_name + ".py")

    with open(outfile_name, "w", encoding="UTF-8") as out:
        out.write(text)


def build(elements, subscript_dict, namespace, outfile_name):
    """
    Constructs and writes the python representation of the model, when the
    the split_modules is set to False in the read_vensim function. The entire
    model is put in a single python file.

    Parameters
    ----------
    elements: list
        Each element is a dictionary, with the various components needed to
        assemble a model component in python syntax. This will contain
        multiple entries for elements that have multiple definitions in the
        original file, and which need to be combined.

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    namespace: dict
        Translation from original model element names (keys) to python safe
        function identifiers (values).

    outfile_name: str
        The name of the file to write the model to.

    Returns
    -------
    None or text: None or str
        If outfile_name="return" it will return the content of the output file
        instead of saving it. It is used for testing.

    """
    # Todo: deal with model level documentation
    # Todo: Make presence of subscript_dict instantiation conditional on usage
    # Todo: Sort elements (alphabetically? group stock funcs?)
    all_elements = merge_partial_elements(elements)
    # separating between control variables and rest of variables
    control_vars_ = [element for element in all_elements if
                     element["py_name"] in ["final_time",
                                            "initial_time",
                                            "saveper",
                                            "time_step"]]
    elements = [element for element in all_elements if element not in
                control_vars_]

    control_vars = _generate_functions(control_vars_, subscript_dict)
    funcs = _generate_functions(elements, subscript_dict)

    text, root = Imports.get_header(os.path.basename(outfile_name))

    text += textwrap.dedent("""
    __pysd_version__ = '%(version)s'

    __data = {
        'scope': None,
        'time': lambda: 0
    }
    %(root)s
    _subscript_dict = %(subscript_dict)s

    _namespace = %(namespace)s
    """ % {
        "subscript_dict": repr(subscript_dict),
        "namespace": repr(namespace),
        "root": root,
        "version": __version__,
    })

    text += _get_control_vars(control_vars) + funcs
    text = black.format_file_contents(text, fast=True, mode=black.FileMode())

    # Needed for various sessions
    build_names.clear()

    # this is used for testing
    if outfile_name == "return":
        return text

    with open(outfile_name, "w", encoding="UTF-8") as out:
        out.write(text)


def _generate_functions(elements, subscript_dict):
    """
    Builds all model elements as functions in string format.
    NOTE: this function calls the build_element function, which updates the
    import_modules.
    Therefore, it needs to be executed before the_generate_automatic_imports
    function.

    Parameters
    ----------
    elements: dict
        Each element is a dictionary, with the various components needed to
        assemble a model component in python syntax. This will contain
        multiple entries for elements that have multiple definitions in the
        original file, and which need to be combined.

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    Returns
    -------
    funcs: str
        String containing all formated model functions

    """
    functions = [build_element(element, subscript_dict) for element in
                 elements]

    funcs = "%(functions)s" % {"functions": "\n".join(functions)}
    funcs = funcs.replace("\t", "    ")

    return funcs


def _get_control_vars(control_vars):
    """
    Create the section of control variables

    Parameters
    ----------
    control_vars: str
        Functions to define control variables.

    Returns
    -------
    text: str
        Control variables section and header of model variables section.

    """
    text = textwrap.dedent("""
    ##########################################################################
    #                            CONTROL VARIABLES                           #
    ##########################################################################

    def _init_outer_references(data):
        for key in data:
            __data[key] = data[key]


    def time():
        return __data['time']()

    """)

    text += control_vars

    text += textwrap.dedent("""
    ##########################################################################
    #                             MODEL VARIABLES                            #
    ##########################################################################
    """)

    return text


def build_element(element, subscript_dict):
    """
    Returns a string that has processed a single element dictionary.

    Parameters
    ----------
    element: dict
        A dictionary containing at least the elements:
        - kind: ['constant', 'setup', 'component', 'lookup']
            Different types of elements will be built differently
        - py_expr: str
            An expression that has been converted already into python syntax
        - subs: list of lists
            Each sublist contains coordinates for initialization of a
            particular part of a subscripted function, the list of
            subscripts vensim attaches to an equation

    subscript_dict: dict
        A dictionary containing the names of subscript families (dimensions)
        as keys, and a list of the possible positions within that dimension
        for each value.

    Returns
    -------
    func: str
        The function to write in the model file.

    """
    if element["kind"] == "constant":
        cache_type = "@cache.run"
    elif element["kind"] in ["component", "component_ext_data"]:
        cache_type = "@cache.step"
    elif element["kind"] == "lookup":
        # lookups may be called with different values in a round
        cache_type = ""
    elif element["kind"] in ["setup", "stateful", "external", "external_add"]:
        # setups only get called once, caching is wasted
        cache_type = ""
    else:
        raise AttributeError("Bad value for 'kind'")

    # check the elements with ADD in their name
    # as these wones are directly added to the
    # external objecets via .add method
    py_expr_no_ADD = ["ADD" not in py_expr for py_expr in element["py_expr"]]

    if sum(py_expr_no_ADD) > 1 and element["kind"] not in [
        "stateful",
        "external",
        "external_add",
    ]:
        py_expr_i = []
        # need to append true to the end as the next element is checked
        py_expr_no_ADD.append(True)
        for i, (py_expr, subs_i) in enumerate(zip(element["py_expr"],
                                                  element["subs"])):
            if not (py_expr.startswith("xr.") or py_expr.startswith("_ext_")):
                # rearrange if it doesn't come from external or xarray
                coords = utils.make_coord_dict(
                    subs_i,
                    subscript_dict,
                    terse=False)
                coords = {
                    new_dim: coords[dim]
                    for new_dim, dim in zip(element["merge_subs"], coords)
                }
                dims = list(coords)
                Imports.add("utils", "rearrange")
                py_expr_i.append("rearrange(%s, %s, %s)" % (
                    py_expr, dims, coords))
            elif py_expr_no_ADD[i]:
                # element comes from external or xarray
                py_expr_i.append(py_expr)
        Imports.add("utils", "xrmerge")
        py_expr = "xrmerge([%s,])" % (
            ",\n".join(py_expr_i))
    else:
        py_expr = element["py_expr"][0]

    contents = "return %s" % py_expr

    element["subs_dec"] = ""
    element["subs_doc"] = "None"

    if element["merge_subs"]:
        # We add the list of the subs to the __doc__ of the function
        # this will give more information to the user and make possible
        # to rewrite subscripted values with model.run(params=X) or
        # model.run(initial_condition=(n,x))
        element["subs_doc"] = "%s" % element["merge_subs"]
        if element["kind"] in ["component", "setup",
                               "constant", "component_ext_data"]:
            # the decorator is not always necessary as the objects
            # defined as xarrays in the model will have the right
            # dimensions always, we should try to reduce to the
            # maximum when we use it
            # re arrange the python object
            element["subs_dec"] =\
                "@subs(%s, _subscript_dict)" % element["merge_subs"]
            Imports.add("subs")

    indent = 8
    element.update(
        {
            "cache": cache_type,
            "ulines": "-" * len(element["real_name"]),
            "contents": contents.replace("\n", "\n" + " " * indent),
        }
    )
    # indent lines 2 onward

    # convert newline indicator and add expected level of indentation
    element["doc"] = element["doc"].replace("\\", "\n").replace("\n", "\n    ")

    if element["kind"] in ["stateful", "external"]:
        func = """
    %(py_name)s = %(py_expr)s
            """ % {
            "py_name": element["py_name"],
            "py_expr": element["py_expr"][0],
        }

    elif element["kind"] == "external_add":
        # external expressions to be added with .add method
        # remove the ADD from the end
        py_name = element["py_name"].split("ADD")[0]
        func = """
    %(py_name)s%(py_expr)s
            """ % {
            "py_name": py_name,
            "py_expr": element["py_expr"][0],
        }

    else:
        sep = "\n" + " " * 10
        if len(element["eqn"]) == 1:
            # Original equation in the same line
            element["eqn"] = element["eqn"][0]
        elif len(element["eqn"]) > 5:
            # First and last original equations separated by vertical dots
            element["eqn"] = (
                sep + element["eqn"][0] + (sep + "  .") * 3 + sep
                    + element["eqn"][-1]
            )
        else:
            # From 2 to 5 equations in different lines
            element["eqn"] = sep + sep.join(element["eqn"])

        func = (
            '''
    %(cache)s
    %(subs_dec)s
    def %(py_name)s(%(arguments)s):
        """
        Real Name: %(real_name)s
        Original Eqn: %(eqn)s
        Units: %(unit)s
        Limits: %(lims)s
        Type: %(kind)s
        Subs: %(subs_doc)s

        %(doc)s
        """
        %(contents)s
        '''
            % element
        )

    return textwrap.dedent(func)


def merge_partial_elements(element_list):
    """
    Merges model elements which collectively all define the model component,
    mostly for multidimensional subscripts

    Parameters
    ----------
    element_list: list
        List of all the elements.

    Returns
    -------
    list:
        List of merged elements.

    """
    outs = dict()  # output data structure

    for element in element_list:
        if element["py_expr"] != "None":  # for
            name = element["py_name"]
            if name not in outs:

                # Use 'expr' for Vensim models, and 'eqn' for Xmile
                # (This makes the Vensim equation prettier.)
                eqn = element["expr"] if "expr" in element else element["eqn"]
                outs[name] = {
                    "py_name": element["py_name"],
                    "real_name": element["real_name"],
                    "doc": element["doc"],
                    "py_expr": [element["py_expr"]],  # in a list
                    "unit": element["unit"],
                    "subs": [element["subs"]],
                    "merge_subs": element["merge_subs"]
                    if "merge_subs" in element else None,
                    "lims": element["lims"],
                    "eqn": [eqn.replace(r"\ ", "")],
                    "kind": element["kind"],
                    "arguments": element["arguments"],
                }

            else:
                eqn = element["expr"] if "expr" in element else element["eqn"]

                outs[name]["doc"] = outs[name]["doc"] or element["doc"]
                outs[name]["unit"] = outs[name]["unit"] or element["unit"]
                outs[name]["lims"] = outs[name]["lims"] or element["lims"]
                outs[name]["eqn"] += [eqn.replace(r"\ ", "")]
                outs[name]["py_expr"] += [element["py_expr"]]
                outs[name]["subs"] += [element["subs"]]
                outs[name]["arguments"] = element["arguments"]

    return list(outs.values())


def add_stock(identifier, expression, initial_condition, subs, merge_subs):
    """
    Creates new model element dictionaries for the model elements associated
    with a stock.

    Parameters
    ----------
    identifier: str
        The python-safe name of the stock.

    expression: str
        The formula which forms the derivative of the stock.

    initial_condition: str
        Formula which forms the initial condition for the stock.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        a string to use in place of the 'INTEG...' pieces in the element
        expression string, a reference to the stateful object.

    new_structure: list
        List of additional model element dictionaries. When there are
        subscripts, constructs an external 'init' and 'ddt' function so
        that these can be appropriately aggregated.

    """
    Imports.add("functions", "Integ")

    new_structure = []
    py_name = "_integ_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = "Integ(lambda: %s, lambda: %s, '%s')" % (
            expression,
            initial_condition,
            py_name,
        )
    else:
        stateful_py_expr = "Integ(_integ_input_%s, _integ_init_%s, '%s')" % (
            identifier,
            identifier,
            py_name,
        )

        # following elements not specified in the model file, but must exist
        # create the stock initialization element
        new_structure.append(
            {
                "py_name": "_integ_init_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "setup",
                "py_expr": initial_condition,
                "subs": subs,
                "merge_subs": merge_subs,
                "doc": "Provides initial conditions for %s function"
                        % identifier,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "arguments": "",
            }
        )

        new_structure.append(
            {
                "py_name": "_integ_input_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "component",
                "doc": "Provides derivative for %s function" % identifier,
                "subs": subs,
                "merge_subs": merge_subs,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "py_expr": expression,
                "arguments": "",
            }
        )

    # describe the stateful object
    new_structure.append(
        {
            "py_name": py_name,
            "parent_name": identifier,
            "real_name": "Representation of  %s" % identifier,
            "doc": "Integrates Expression %s" % expression,
            "py_expr": stateful_py_expr,
            "unit": "None",
            "lims": "None",
            "eqn": "None",
            "subs": "",
            "merge_subs": None,
            "kind": "stateful",
            "arguments": "",
        }
    )

    return "%s()" % py_name, new_structure


def add_delay(identifier, delay_input, delay_time, initial_value, order,
              subs, merge_subs):
    """
    Creates code to instantiate a stateful 'Delay' object,
    and provides reference to that object's output.

    The name of the stateful object is based upon the passed in parameters,
    so if there are multiple places where identical delay functions are
    referenced, the translated python file will only maintain one stateful
    object, and reference it multiple times.

    Parameters
    ----------
    identifier: str
        The python-safe name of the delay.

    delay_input: str
        Reference to the model component that is the input to the delay.

    delay_time: str
        Can be a number (in string format) or a reference to another model
        element which will calculate the delay. This is calculated throughout
        the simulation at runtime.

    initial_value: str
        This is used to initialize the stocks that are present in the delay.
        We initialize the stocks with equal values so that the outflow in
        the first timestep is equal to this value.

    order: str
        The number of stocks in the delay pipeline. As we construct the
        delays at build time, this must be an integer and cannot be calculated
        from other model components. Anything else will yield a ValueError.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        Reference to the delay object `__call__` method, which will return
        the output of the delay process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "Delay")

    new_structure = []
    py_name = "_delay_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = (
            "Delay(lambda: %s, lambda: %s,"
            "lambda: %s, lambda: %s, time_step, '%s')"
            % (delay_input, delay_time, initial_value, order, py_name)
        )

    else:
        stateful_py_expr = (
            "Delay(_delay_input_%s, lambda: %s, _delay_init_%s,"
            "lambda: %s, time_step, '%s')"
            % (identifier, delay_time, identifier, order, py_name)
        )

        # following elements not specified in the model file, but must exist
        # create the delay initialization element
        new_structure.append(
            {
                "py_name": "_delay_init_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "setup",  # not specified in the model file, but must
                # exist
                "py_expr": initial_value,
                "subs": subs,
                "merge_subs": merge_subs,
                "doc": "Provides initial conditions for %s function" \
                        % identifier,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "arguments": "",
            }
        )

        new_structure.append(
            {
                "py_name": "_delay_input_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "component",
                "doc": "Provides input for %s function" % identifier,
                "subs": subs,
                "merge_subs": merge_subs,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "py_expr": delay_input,
                "arguments": "",
            }
        )

    # describe the stateful object
    new_structure.append(
        {
            "py_name": py_name,
            "parent_name": identifier,
            "real_name": "Delay of %s" % delay_input,
            "doc": "Delay time: %s \n Delay initial value %s \n Delay order %s"
            % (delay_time, initial_value, order),
            "py_expr": stateful_py_expr,
            "unit": "None",
            "lims": "None",
            "eqn": "None",
            "subs": "",
            "merge_subs": None,
            "kind": "stateful",
            "arguments": "",
        }
    )

    return "%s()" % py_name, new_structure


def add_delay_f(identifier, delay_input, delay_time, initial_value):
    """
    Creates code to instantiate a stateful 'DelayFixed' object,
    and provides reference to that object's output.

    The name of the stateful object is based upon the passed in parameters,
    so if there are multiple places where identical delay functions are
    referenced, the translated python file will only maintain one stateful
    object, and reference it multiple times.

    Parameters
    ----------
    identifier: str
        The python-safe name of the delay.

    delay_input: str
        Reference to the model component that is the input to the delay.

    delay_time: str
        Can be a number (in string format) or a reference to another model
        element which will calculate the delay. This is calculated throughout
        the simulation at runtime.

    initial_value: str
        This is used to initialize the stocks that are present in the delay.
        We initialize the stocks with equal values so that the outflow in
        the first timestep is equal to this value.

    Returns
    -------
    reference: str
        Reference to the delay object `__call__` method, which will return
        the output of the delay process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "DelayFixed")

    py_name = "_delayfixed_%s" % identifier

    stateful_py_expr = (
        "DelayFixed(lambda: %s, lambda: %s,"
        "lambda: %s, time_step, '%s')"
        % (delay_input, delay_time, initial_value, py_name)
    )

    # describe the stateful object
    stateful = {
        "py_name": py_name,
        "parent_name": identifier,
        "real_name": "Delay fixed  of %s" % delay_input,
        "doc": "DelayFixed time: %s \n Delay initial value %s"
        % (delay_time, initial_value),
        "py_expr": stateful_py_expr,
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": "",
        "merge_subs": None,
        "kind": "stateful",
        "arguments": "",
    }

    return "%s()" % py_name, [stateful]


def add_n_delay(identifier, delay_input, delay_time, initial_value, order,
                subs, merge_subs):
    """
    Creates code to instantiate a stateful 'DelayN' object,
    and provides reference to that object's output.

    The name of the stateful object is based upon the passed in parameters,
    so if there are multiple places where identical delay functions are
    referenced, the translated python file will only maintain one stateful
    object, and reference it multiple times.

    Parameters
    ----------
    identifier: str
        The python-safe name of the delay.

    delay_input: str
        Reference to the model component that is the input to the delay.

    delay_time: str
        Can be a number (in string format) or a reference to another model
        element which will calculate the delay. This is calculated throughout
        the simulation at runtime.

    initial_value: str
        This is used to initialize the stocks that are present in the delay.
        We initialize the stocks with equal values so that the outflow in
        the first timestep is equal to this value.

    order: str
        The number of stocks in the delay pipeline. As we construct the
        delays at build time, this must be an integer and cannot be calculated
        from other model components. Anything else will yield a ValueError.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        Reference to the delay object `__call__` method, which will return
        the output of the delay process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "DelayN")

    new_structure = []
    py_name = "_delayn_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = (
            "DelayN(lambda: %s, lambda: %s,"
            "lambda: %s, lambda: %s, time_step, '%s')"
            % (delay_input, delay_time, initial_value, order, py_name)
        )

    else:
        stateful_py_expr = (
            "DelayN(_delayn_input_%s, lambda: %s,"
            " _delayn_init_%s, lambda: %s, time_step, '%s')"
            % (identifier, delay_time, identifier, order, py_name)
        )

        # following elements not specified in the model file, but must exist
        # create the delay initialization element
        new_structure.append(
            {
                "py_name": "_delayn_init_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "setup",  # not specified in the model file, but must
                # exist
                "py_expr": initial_value,
                "subs": subs,
                "merge_subs": merge_subs,
                "doc": "Provides initial conditions for %s function" \
                        % identifier,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "arguments": "",
            }
        )

        new_structure.append(
            {
                "py_name": "_delayn_input_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "component",
                "doc": "Provides input for %s function" % identifier,
                "subs": subs,
                "merge_subs": merge_subs,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "py_expr": delay_input,
                "arguments": "",
            }
        )

    # describe the stateful object
    new_structure.append(
        {
            "py_name": py_name,
            "parent_name": identifier,
            "real_name": "DelayN of %s" % delay_input,
            "doc": "DelayN time: %s \n DelayN initial value %s \n DelayN order\
                    %s"
            % (delay_time, initial_value, order),
            "py_expr": stateful_py_expr,
            "unit": "None",
            "lims": "None",
            "eqn": "None",
            "subs": "",
            "merge_subs": None,
            "kind": "stateful",
            "arguments": "",
        }
    )

    return "%s()" % py_name, new_structure


def add_sample_if_true(identifier, condition, actual_value, initial_value,
                       subs, merge_subs):
    """
    Creates code to instantiate a stateful 'SampleIfTrue' object,
    and provides reference to that object's output.

    Parameters
    ----------
    identifier: str
        The python-safe name of the sample if true.

    condition: str
        Reference to another model element that is the condition to the
        'sample if true' function.

    actual_value: str
        Can be a number (in string format) or a reference to another model
        element which is calculated throughout the simulation at runtime.

    initial_value: str
        This is used to initialize the state of the sample if true function.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        Reference to the sample if true object `__call__` method,
        which will return the output of the sample if true process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "SampleIfTrue")

    new_structure = []
    py_name = "_sample_if_true_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = "SampleIfTrue(lambda: %s, lambda: %s,"\
                           "lambda: %s, '%s')" % (
                               condition, actual_value, initial_value, py_name)

    else:
        stateful_py_expr = "SampleIfTrue(lambda: %s, lambda: %s,"\
                           "_sample_if_true_init_%s, '%s')" % (
                               condition, actual_value, identifier, py_name)

        # following elements not specified in the model file, but must exist
        # create the delay initialization element
        new_structure.append({
            "py_name": "_sample_if_true_init_%s" % identifier,
            "parent_name": identifier,
            "real_name": "Implicit",
            "kind": "setup",  # not specified in the model file, but must exist
            "py_expr": initial_value,
            "subs": subs,
            "merge_subs": merge_subs,
            "doc": "Provides initial conditions for %s function" % identifier,
            "unit": "See docs for %s" % identifier,
            "lims": "None",
            "eqn": "None",
            "arguments": ""
        })
    # describe the stateful object
    new_structure.append({
        "py_name": py_name,
        "parent_name": identifier,
        "real_name": "Sample if true of %s" % identifier,
        "doc": "Initial value: %s \n  Input: %s \n Condition: %s" % (
            initial_value, actual_value, condition),
        "py_expr": stateful_py_expr,
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": "",
        "merge_subs": None,
        "kind": "stateful",
        "arguments": ""
    })

    return "%s()" % py_name, new_structure


def add_n_smooth(identifier, smooth_input, smooth_time, initial_value, order,
                 subs, merge_subs):
    """
    Constructs stock and flow chains that implement the calculation of
    a smoothing function.

    Parameters
    ----------
    identifier: str
        The python-safe name of the smooth.

    smooth_input: str
        Reference to the model component that is the input to the
        smoothing function.

    smooth_time: str
        Can be a number (in string format) or a reference to another model
        element which will calculate the delay. This is calculated throughout
        the simulation at runtime.

    initial_value: str
        This is used to initialize the stocks that are present in the delay.
        We initialize the stocks with equal values so that the outflow in
        the first timestep is equal to this value.

    order: str
        The number of stocks in the delay pipeline. As we construct the delays
        at build time, this must be an integer and cannot be calculated from
        other model components. Anything else will yield a ValueError.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        Reference to the smooth object `__call__` method, which will return
        the output of the smooth process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "Smooth")

    new_structure = []
    py_name = "_smooth_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = (
            "Smooth(lambda: %s, lambda: %s,"
            "lambda: %s, lambda: %s, '%s')"
            % (smooth_input, smooth_time, initial_value, order, py_name)
        )

    else:
        # only need to re-dimension init and input as xarray will take care of
        # other
        stateful_py_expr = (
            "Smooth(_smooth_input_%s, lambda: %s,"
            " _smooth_init_%s, lambda: %s, '%s')"
            % (identifier, smooth_time, identifier, order, py_name)
        )

        # following elements not specified in the model file, but must exist
        # create the delay initialization element
        new_structure.append(
            {
                "py_name": "_smooth_init_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "setup",  # not specified in the model file, but must
                # exist
                "py_expr": initial_value,
                "subs": subs,
                "merge_subs": merge_subs,
                "doc": "Provides initial conditions for %s function" % \
                       identifier,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "arguments": "",
            }
        )

        new_structure.append(
            {
                "py_name": "_smooth_input_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "component",
                "doc": "Provides input for %s function" % identifier,
                "subs": subs,
                "merge_subs": merge_subs,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "py_expr": smooth_input,
                "arguments": "",
            }
        )

    new_structure.append(
        {
            "py_name": py_name,
            "parent_name": identifier,
            "real_name": "Smooth of %s" % smooth_input,
            "doc": "Smooth time:" +
                   "%s \n Smooth initial value %s \n Smooth order %s"
                   % (smooth_time, initial_value, order),
            "py_expr": stateful_py_expr,
            "unit": "None",
            "lims": "None",
            "eqn": "None",
            "subs": "",
            "merge_subs": None,
            "kind": "stateful",
            "arguments": "",
        }
    )

    return "%s()" % py_name, new_structure


def add_n_trend(identifier, trend_input, average_time, initial_trend,
                subs, merge_subs):
    """
    Constructs Trend object.

    Parameters
    ----------
    identifier: str
        The python-safe name of the trend.

    trend_input: str
        Input of the trend.

    average_time: str
        Average time of the trend.

    trend_initial: str
        This is used to initialize the trend .

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    reference: str
        Reference to the trend object `__call__` method, which will return the
        output of the trend process.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "Trend")

    new_structure = []
    py_name = "_trend_%s" % identifier

    if len(subs) == 0:
        stateful_py_expr = "Trend(lambda: %s, lambda: %s,"\
                           " lambda: %s, '%s')" % (
                               trend_input, average_time,
                               initial_trend, py_name)

    else:
        # only need to re-dimension init as xarray will take care of other
        stateful_py_expr = "Trend(lambda: %s, lambda: %s,"\
                           " _trend_init_%s, '%s')" % (
                               trend_input, average_time,
                               identifier, py_name)

        # following elements not specified in the model file, but must exist
        # create the delay initialization element
        new_structure.append(
            {
                "py_name": "_trend_init_%s" % identifier,
                "parent_name": identifier,
                "real_name": "Implicit",
                "kind": "setup",  # not specified in the model file, but must
                # exist
                "py_expr": initial_trend,
                "subs": subs,
                "merge_subs": merge_subs,
                "doc": "Provides initial conditions for %s function"
                        % identifier,
                "unit": "See docs for %s" % identifier,
                "lims": "None",
                "eqn": "None",
                "arguments": "",
            }
        )

    new_structure.append(
        {
            "py_name": py_name,
            "parent_name": identifier,
            "real_name": "trend of %s" % trend_input,
            "doc": "Trend average time: %s \n Trend initial value %s"
            % (average_time, initial_trend),
            "py_expr": stateful_py_expr,
            "unit": "None",
            "lims": "None",
            "eqn": "None",
            "subs": "",
            "merge_subs": None,
            "kind": "stateful",
            "arguments": "",
        }
    )

    return "%s()" % py_name, new_structure


def add_initial(identifier, value):
    """
    Constructs a stateful object for handling vensim's 'Initial' functionality.

    Parameters
    ----------
    identifier: str
        The python-safe name of the initial.

    value: str
        The expression which will be evaluated, and the first value of
        which returned.

    Returns
    -------
    reference: str
        Reference to the Initial object `__call__` method,
        which will return the first calculated value of `identifier`.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("functions", "Initial")
    py_name = utils.make_python_identifier("_initial_%s"
                                           % identifier)[0]

    stateful = {
        "py_name": py_name,
        "parent_name": identifier,
        "real_name": "Initial %s" % identifier,
        "doc": "Returns the value taken on during the initialization phase",
        "py_expr": "Initial(lambda: %s, '%s')" % (value, py_name),
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": "",
        "merge_subs": None,
        "kind": "stateful",
        "arguments": "",
    }

    return "%s()" % stateful["py_name"], [stateful]


def add_ext_data(identifier, file_name, tab, time_row_or_col, cell, subs,
                 subscript_dict, merge_subs, keyword):
    """
    Constructs a external object for handling Vensim's GET XLS DATA and
    GET DIRECT DATA functionality.

    Parameters
    ----------
    identifier: str
        The python-safe name of the external values.

    file_name: str
        Filepath to the data.

    tab: str
        Tab where the data is.

    time_row_or_col: str
        Identifier to the starting point of the time dimension.

    cell: str
        Cell identifier where the data starts.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    subscript_dict: dict
        Dictionary describing the possible dimensions of the stock's
        subscripts.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects.

    keyword: str
        Data retrieval method ('interpolate', 'look forward', 'hold backward').

    Returns
    -------
    reference: str
        Reference to the ExtData object `__call__` method, which will
        return the retrieved value of data for the current time step.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("external", "ExtData")

    coords = utils.simplify_subscript_input(
        utils.make_coord_dict(subs, subscript_dict, terse=False),
        subscript_dict, return_full=False, merge_subs=merge_subs)
    keyword = (
        "'%s'" % keyword.strip(":").lower() if isinstance(keyword, str) else
        keyword)
    name = utils.make_python_identifier("_ext_data_%s" % identifier)[0]

    # Check if the object already exists
    if name in build_names:
        # Create a new py_name with ADD_# ending
        # This object name will not be used in the model as
        # the information is added to the existing object
        # with add method.
        kind = "external_add"
        name = utils.make_add_identifier(name, build_names)
        py_expr = ".add(%s, %s, %s, %s, %s, %s)"
    else:
        # Regular name will be used and a new object will be created
        # in the model file.
        build_names.add(name)
        kind = "external"
        py_expr = "ExtData(%s, %s, %s, %s, %s, %s,\n"\
                  "        _root, '{}')".format(name)

    external = {
        "py_name": name,
        "parent_name": identifier,
        "real_name": "External data for %s" % identifier,
        "doc": "Provides data for data variable %s" % identifier,
        "py_expr": py_expr % (file_name, tab, time_row_or_col, cell, keyword,
                              coords),
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": subs,
        "merge_subs": merge_subs,
        "kind": kind,
        "arguments": "",
    }

    return "%s(time())" % external["py_name"], [external]


def add_ext_constant(identifier, file_name, tab, cell,
                     subs, subscript_dict, merge_subs):
    """
    Constructs a external object for handling Vensim's GET XLS CONSTANT and
    GET DIRECT CONSTANT functionality.

    Parameters
    ----------
    identifier: str
        The python-safe name of the external values.

    file_name: str
        Filepath to the data.

    tab: str
        Tab where the data is.

    cell: str
        Cell identifier where the data starts.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    subscript_dict: dict
        Dictionary describing the possible dimensions of the stock's
        subscripts.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects.

    Returns
    -------
    reference: str
        Reference to the ExtConstant object `__call__` method,
        which will return the read value of the data.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("external", "ExtConstant")

    coords = utils.simplify_subscript_input(
        utils.make_coord_dict(subs, subscript_dict, terse=False),
        subscript_dict, return_full=False, merge_subs=merge_subs)
    name = utils.make_python_identifier("_ext_constant_%s" % identifier)[0]

    # Check if the object already exists
    if name in build_names:
        # Create a new py_name with ADD_# ending
        # This object name will not be used in the model as
        # the information is added to the existing object
        # with add method.
        kind = "external_add"
        name = utils.make_add_identifier(name, build_names)
        py_expr = ".add(%s, %s, %s, %s)"
    else:
        # Regular name will be used and a new object will be created
        # in the model file.
        kind = "external"
        py_expr = "ExtConstant(%s, %s, %s, %s,\n"\
                  "            _root, '{}')".format(name)
    build_names.add(name)

    external = {
        "py_name": name,
        "parent_name": identifier,
        "real_name": "External constant for %s" % identifier,
        "doc": "Provides data for constant data variable %s" % identifier,
        "py_expr": py_expr % (file_name, tab, cell, coords),
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": subs,
        "merge_subs": merge_subs,
        "kind": kind,
        "arguments": "",
    }

    return "%s()" % external["py_name"], [external]


def add_ext_lookup(identifier, file_name, tab, x_row_or_col, cell,
                   subs, subscript_dict, merge_subs):
    """
    Constructs a external object for handling Vensim's GET XLS LOOKUPS and
    GET DIRECT LOOKUPS functionality.

    Parameters
    ----------
    identifier: str
        The python-safe name of the external values.

    file_name: str
        Filepath to the data.

    tab: str
        Tab where the data is.

    x_row_or_col: str
        Identifier to the starting point of the lookup dimension.

    cell: str
        Cell identifier where the data starts.

    subs: list of strings
        List of strings of subscript indices that correspond to the
        list of expressions, and collectively define the shape of the output.

    subscript_dict: dict
        Dictionary describing the possible dimensions of the stock's
        subscripts.

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects.

    Returns
    -------
    reference: str
        Reference to the ExtLookup object `__call__` method,
        which will return the retrieved value of data after interpolating it.

    new_structure: list
        List of element construction dictionaries for the builder to assemble.

    """
    Imports.add("external", "ExtLookup")

    coords = utils.simplify_subscript_input(
        utils.make_coord_dict(subs, subscript_dict, terse=False),
        subscript_dict, return_full=False, merge_subs=merge_subs)
    name = utils.make_python_identifier("_ext_lookup_%s" % identifier)[0]

    # Check if the object already exists
    if name in build_names:
        # Create a new py_name with ADD_# ending
        # This object name will not be used in the model as
        # the information is added to the existing object
        # with add method.
        kind = "external_add"
        name = utils.make_add_identifier(name, build_names)
        py_expr = ".add(%s, %s, %s, %s, %s)"
    else:
        # Regular name will be used and a new object will be created
        # in the model file.
        kind = "external"
        py_expr = "ExtLookup(%s, %s, %s, %s, %s,\n"\
                  "          _root, '{}')".format(name)
    build_names.add(name)

    external = {
        "py_name": name,
        "parent_name": identifier,
        "real_name": "External lookup data for %s" % identifier,
        "doc": "Provides data for external lookup variable %s" % identifier,
        "py_expr": py_expr % (file_name, tab, x_row_or_col, cell, coords),
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": subs,
        "merge_subs": merge_subs,
        "kind": kind,
        "arguments": "x",
    }

    return "%s(x)" % external["py_name"], [external]


def add_macro(macro_name, filename, arg_names, arg_vals):
    """
    Constructs a stateful object instantiating a 'Macro'

    Parameters
    ----------
    macro_name: str
        Python safe name for macro.

    filename: str
        Filepath to macro definition.

    func_args: dict
        Dictionary of values to be passed to macro, {key: function}.

    Returns
    -------
    reference: str
        reference to the Initial object `__call__` method,
        which will return the first calculated value of `initial_input`
    new_structure: list
        list of element construction dictionaries for the builder to assemble

    """
    Imports.add("functions", "Macro")

    py_name = "_macro_" + macro_name + "_" + "_".join(
        [utils.make_python_identifier(f)[0] for f in arg_vals])

    func_args = "{ %s }" % ", ".join(
        ["'%s': lambda: %s" % (key, val) for key, val in zip(arg_names,
                                                             arg_vals)])

    stateful = {
        "py_name": py_name,
        "parent_name": macro_name,
        "real_name": "Macro Instantiation of " + macro_name,
        "doc": "Instantiates the Macro",
        "py_expr": "Macro('%s', %s, '%s',"
        " time_initialization=lambda: __data['time'],"
        " py_name='%s')" % (filename, func_args, macro_name, py_name),
        "unit": "None",
        "lims": "None",
        "eqn": "None",
        "subs": "",
        "merge_subs": None,
        "kind": "stateful",
        "arguments": "",
    }

    return "%s()" % stateful["py_name"], [stateful]


def add_incomplete(var_name, dependencies):
    """
    Incomplete functions don't really need to be 'builders' as they
    add no new real structure, but it's helpful to have a function
    in which we can raise a warning about the incomplete equation
    at translate time.

    parameters
    ----------
    var_name: str
        The python-safe name of the incomplete variable.

    dependencies: list
        The list of the dependencies in the variable.

    Returns
    -------
    str:
        Inclompete funcion call.

    """
    Imports.add("functions", "incomplete")

    warnings.warn(
        "%s has no equation specified" % var_name, SyntaxWarning, stacklevel=2
    )

    # first arg is `self` reference
    return "incomplete(%s)" % ", ".join(dependencies), []


def build_function_call(function_def, user_arguments):
    """

    Parameters
    ----------
    function_def: dict
        Function definition map with following keys:
        - name: name of the function.
        - parameters: list with description of all parameters of this function
            - name
            - optional?
            - type: [
                "expression", - provide converted expression as parameter for
                                runtime evaluating before the method call
                "lambda",     - provide lambda expression as parameter for
                                delayed runtime evaluation in the method call
                "time",       - provide access to current instance of
                                time object
                "scope"       - provide access to current instance of
                                scope object (instance of Macro object)
            ]

    user_arguments: list of arguments provided from model.

    Returns
    -------
    str:
        Function call.

    """
    if isinstance(function_def, str):
        return function_def + "(" + ",".join(user_arguments) + ")"

    if function_def["name"] == "not_implemented_function":
        user_arguments = ["'" + function_def["original_name"] + "'"] + \
            user_arguments
        warnings.warn(
            "\n\nTrying to translate "
            + function_def["original_name"]
            + " which it is not implemented on PySD. The translated "
            + "model will crash... "
        )

    if "module" in function_def:
        if function_def["module"] in ["numpy", "xarray"]:
            # import external modules
            Imports.add(function_def["module"])
        else:
            # import method from PySD module
            Imports.add(function_def["module"], function_def["name"])

    if "parameters" in function_def:
        parameters = function_def["parameters"]
        arguments = []
        argument_idx = 0
        for parameter_idx in range(len(parameters)):
            parameter_def = parameters[parameter_idx]
            is_optional = (
                parameter_def["optional"] if "optional" in parameter_def else
                False
            )
            if argument_idx >= len(user_arguments) and is_optional:
                break

            parameter_type = (
                parameter_def["type"] if "type" in parameter_def else
                "expression")

            user_argument = user_arguments[argument_idx]
            if parameter_type in ["expression", "lambda"]:
                argument_idx += 1

            arguments.append(
                {
                    "expression": user_argument,
                    "lambda": "lambda: " + user_argument,
                    "time": "__data['time']",
                    "scope": "__data['scope']",
                }[parameter_type]
            )

        return function_def["name"] + "(" + ", ".join(arguments) + ")"

    return function_def["name"] + "(" + ",".join(user_arguments) + ")"



"""
These are general utilities used by the builder.py, functions.py or the
model file. Vensim's function equivalents should not go here but in
functions.py
"""

import os
import warnings
import keyword
import json

import regex as re
import progressbar
import numpy as np
import xarray as xr

# used to create python safe names
from .decorators import __dir__ as ddir
from .external import __dir__ as edir
from .functions import __dir__ as fdir


def xrmerge(das, accept_new=True):
    """
    Merges xarrays with different dimension sets.

    Parameters
    ----------
    das: list of xarray.DataArrays
        The list of the data arrays to merge.

    accept_new: bool (optional)
        Default is True.

    Returns
    -------
    da: xarray.DataArray
        Merged data array.

    References
    ----------
    Thanks to @jcmgray
    https://github.com/pydata/xarray/issues/742#issue-130753818

    In the future, we may not need this as xarray may provide the merge for us.
    """
    da = das[0]
    for new_da in das[1:]:
        # Expand both to have same dimensions, padding with NaN
        da, new_da = xr.align(da, new_da, join="outer")
        # Fill NaNs one way or the other re. accept_new
        da = new_da.fillna(da) if accept_new else da.fillna(new_da)
    return da


def xrsplit(array):
    """
    Split an array to a list of all the components.

    Parameters
    ----------
    array: xarray.DataArray
        Array to split.

    Returns
    -------
    sp_list: list of xarray.DataArrays
        List of shape 0 xarray.DataArrays with coordinates.

    """
    sp_list = [sa for sa in array]
    if sp_list[0].shape:
        sp_list = [ssa for sa in sp_list for ssa in xrsplit(sa)]
    return sp_list


def find_subscript_name(subscript_dict, element, avoid=[]):
    """
    Given a subscript dictionary, and a member of a subscript family,
    return the first key of which the member is within the value list.
    If element is already a subscript name, return that.

    Parameters
    ----------
    subscript_dict: dict
        Follows the {'subscript name':['list','of','subscript','elements']}
        format.

    element: str

    avoid: list (optional)
        List of subscripts to avoid. Default is an empty list.

    Returns
    -------

    Examples
    --------
    >>> find_subscript_name({'Dim1': ['A', 'B'],
    ...                      'Dim2': ['C', 'D', 'E'],
    ...                      'Dim3': ['F', 'G', 'H', 'I']},
    ...                      'D')
    'Dim2'
    >>> find_subscript_name({'Dim1': ['A', 'B'],
    ...                      'Dim2': ['A', 'B'],
    ...                      'Dim3': ['A', 'B']},
    ...                      'B')
    'Dim1'
    >>> find_subscript_name({'Dim1': ['A', 'B'],
    ...                      'Dim2': ['A', 'B'],
    ...                      'Dim3': ['A', 'B']},
    ...                      'B',
    ...                      avoid=['Dim1'])
    'Dim2'
    """
    if element in subscript_dict.keys():
        return element

    for name, elements in subscript_dict.items():
        if element in elements and name not in avoid:
            return name


def make_coord_dict(subs, subscript_dict, terse=True):
    """
    This is for assisting with the lookup of a particular element, such that
    the output of this function would take the place of %s in this expression.

    `variable.loc[%s]`

    Parameters
    ----------
    subs: list of strings
        coordinates, either as names of dimensions, or positions within
        a dimension.

    subscript_dict: dict
        the full dictionary of subscript names and values.

    terse: bool (optional)
        If True, includes only elements that do not cover the full range of
        values in their respective dimension.If False, returns all dimensions.
        Default is True.

    Returns
    -------
    coordinates: dict
        Coordinates needed to access the xarray quantities we're interested in.

    Examples
    --------
    >>> make_coord_dict(['Dim1', 'D'], {'Dim1': ['A', 'B', 'C'],
    ...                                 'Dim2': ['D', 'E', 'F']})
    {'Dim2': ['D']}
    >>> make_coord_dict(['Dim1', 'D'], {'Dim1': ['A', 'B', 'C'],
    ...                                 'Dim2':['D', 'E', 'F']}, terse=False)
    {'Dim2': ['D'], 'Dim1': ['A', 'B', 'C']}

    """
    sub_elems_list = [y for x in subscript_dict.values() for y in x]
    coordinates = {}
    for sub in subs:
        if sub in sub_elems_list:
            name = find_subscript_name(subscript_dict, sub, avoid=subs)
            coordinates[name] = [sub]
        elif not terse:
            coordinates[sub] = subscript_dict[sub]
    return coordinates


def make_merge_list(subs_list, subscript_dict, element=""):
    """
    This is for assisting when building xrmerge. From a list of subscript
    lists returns the final subscript list after mergin. Necessary when
    merging variables with subscripts comming from different definitions.

    Parameters
    ----------
    subs_list: list of lists of strings
        Coordinates, either as names of dimensions, or positions within
        a dimension.

    subscript_dict: dict
        The full dictionary of subscript names and values.

    element: str (optional)
        Element name, if given it will be printed with any error or
        warning message. Default is "".

    Returns
    -------
    dims: list
        Final subscripts after merging.

    Examples
    --------
    >>> make_merge_list([['upper'], ['C']], {'all': ['A', 'B', 'C'],
    ...                                      'upper': ['A', 'B']})
    ['all']

    """
    coords_set = [set() for i in range(len(subs_list[0]))]
    coords_list = [
        make_coord_dict(subs, subscript_dict, terse=False)
        for subs in subs_list
    ]

    # update coords set
    [[coords_set[i].update(coords[dim]) for i, dim in enumerate(coords)]
     for coords in coords_list]

    dims = [None] * len(coords_set)
    # create an array with the name of the subranges for all merging elements
    dims_list = np.array([list(coords) for coords in coords_list]).transpose()
    indexes = np.arange(len(dims))

    for i, coord2 in enumerate(coords_set):
        dims1 = [
            dim for dim in dims_list[i]
            if dim is not None and set(subscript_dict[dim]) == coord2
        ]
        if dims1:
            # if the given coordinate already matches return it
            dims[i] = dims1[0]
        else:
            # find a suitable coordinate
            other_dims = dims_list[indexes != i]
            for name, elements in subscript_dict.items():
                if coord2 == set(elements) and name not in other_dims:
                    dims[i] = name
                    break

            if not dims[i]:
                # the dimension is incomplete use the smaller
                # dimension that completes it
                for name, elements in subscript_dict.items():
                    if coord2.issubset(set(elements))\
                      and name not in other_dims:
                        dims[i] = name
                        warnings.warn(
                            element
                            + "\nDimension given by subscripts:"
                            + "\n\t{}\nis incomplete ".format(coord2)
                            + "using {} instead.".format(name)
                            + "\nSubscript_dict:"
                            + "\n\t{}".format(subscript_dict)
                        )
                        break

            if not dims[i]:
                for name, elements in subscript_dict.items():
                    if coord2 == set(elements):
                        j = 1
                        while name + str(j) in subscript_dict.keys():
                            j += 1
                        subscript_dict[name + str(j)] = elements
                        dims[i] = name + str(j)
                        warnings.warn(
                            element
                            + "\nAdding new subscript range to"
                            + " subscript_dict:\n"
                            + name + str(j) + ": " + ', '.join(elements))
                        break

            if not dims[i]:
                # not able to find the correct dimension
                raise ValueError(
                    element
                    + "\nImpossible to find the dimension that contains:"
                    + "\n\t{}\nFor subscript_dict:".format(coord2)
                    + "\n\t{}".format(subscript_dict)
                )

    return dims


def make_python_identifier(
    string, namespace=None, reserved_words=None, convert="drop", handle="force"
):
    """
    Takes an arbitrary string and creates a valid Python identifier.

    If the input string is in the namespace, return its value.

    If the python identifier created is already in the namespace,
    but the input string is not (ie, two similar strings resolve to
    the same python identifier)

    or if the identifier is a reserved word in the reserved_words
    list, or is a python default reserved word,
    adds _1, or if _1 is in the namespace, _2, etc.

    Parameters
    ----------
    string: str
        The text to be converted into a valid python identifier.

    namespace: dict
        Map of existing translations into python safe identifiers.
        This is to ensure that two strings are not translated into
        the same python identifier.

    reserved_words: list of strings (optional)
        List of words that are reserved (because they have other meanings
        in this particular program, such as also being the names of
        libraries, etc.

    convert: str (optional)
        Tells the function what to do with characters that are not
        valid in python identifiers.
        - 'hex' implies that they will be converted to their hexidecimal
                representation. This is handy if you have variables that
                have a lot of reserved characters, or you don't want the
                name to be dependent on when things were added to the
                namespace.
        - 'drop' implies that they will just be dropped altogether.

    handle: str (optional)
        Tells the function how to deal with namespace conflicts
        - 'force' will create a representation which is not in conflict
                  by appending _n to the resulting variable where n is
                  the lowest number necessary to avoid a conflict
        - 'throw' will raise an exception

    Returns
    -------
    identifier: str
        A vaild python identifier based on the input string

    namespace: dict
        An updated map of the translations of words to python identifiers,
        including the passed in 'string'.

    Examples
    --------
    >>> make_python_identifier('Capital')
    ('capital', {'Capital': 'capital'})

    >>> make_python_identifier('multiple words')
    ('multiple_words', {'multiple words': 'multiple_words'})

    >>> make_python_identifier('multiple     spaces')
    ('multiple_spaces', {'multiple     spaces': 'multiple_spaces'})

    When the name is a python keyword, add '_1' to differentiate it
    >>> make_python_identifier('for')
    ('for_1', {'for': 'for_1'})

    Remove leading and trailing whitespace
    >>> make_python_identifier('  whitespace  ')
    ('whitespace', {'  whitespace  ': 'whitespace'})

    Remove most special characters outright:
    >>> make_python_identifier('H@t tr!ck')
    ('ht_trck', {'H@t tr!ck': 'ht_trck'})

    Replace special characters with their hex representations
    >>> make_python_identifier('H@t tr!ck', convert='hex')
    ('h40t_tr21ck', {'H@t tr!ck': 'h40t_tr21ck'})

    remove leading digits
    >>> make_python_identifier('123abc')
    ('abc', {'123abc': 'abc'})

    already in namespace
    >>> make_python_identifier('Var$', namespace={'Var$': 'var'})
    ('var', {'Var$': 'var'})

    namespace conflicts
    >>> make_python_identifier('Var@', namespace={'Var$': 'var'})
    ('var_1', {'Var$': 'var', 'Var@': 'var_1'})

    >>> make_python_identifier('Var$', namespace={'Var@': 'var',
    ...                                           'Var%':'var_1'})
    ('var_2', {'Var@': 'var', 'Var%': 'var_1', 'Var$': 'var_2'})

    throw exception instead
    >>> make_python_identifier('Var$', namespace={'Var@': 'var'},
    ...                        handle='throw')
    Traceback (most recent call last):
     ...
    NameError: variable already exists in namespace or is a reserved word


    References
    ----------
    Identifiers must follow the convention outlined here:
        https://docs.python.org/2/reference/lexical_analysis.html#identifiers

    """
    if namespace is None:
        namespace = dict()

    if reserved_words is None:
        reserved_words = list()

    # reserved the names of PySD functions and methods
    reserved_words += dir() + fdir() + edir() + ddir()

    if string in namespace:
        return namespace[string], namespace

    # create a working copy (and make it lowercase, while we're at it)
    s = string.lower()

    # remove leading and trailing whitespace
    s = s.strip()

    # Make spaces into underscores
    s = re.sub(r"[\s\t\n]+", "_", s)

    if convert == "hex":
        # Convert invalid characters to hex. Note: \p{l} designates all
        # Unicode letter characters (any language), \p{m} designates all
        # mark symbols (e.g., vowel marks in Indian scrips, such as the final)
        # and \p{n} designates all numbers. We allow any of these to be
        # present in the regex.
        s = "".join(
            [c.encode("hex") if re.findall(r"[^\p{l}\p{m}\p{n}_]", c)
             else c for c in s]
        )

    elif convert == "drop":
        # Remove invalid characters
        s = re.sub(r"[^\p{l}\p{m}\p{n}_]", "", s)

    # If leading characters are not a letter or underscore add nvs_.
    # Only letters can be leading characters.
    if re.findall(r"^[^\p{l}_]+", s):
        s = "nvs_" + s

    # Check that the string is not a python identifier
    while (s in keyword.kwlist or
           s in namespace.values() or
           s in reserved_words):
        if handle == "throw":
            raise NameError(
                s + " already exists in namespace or is a reserved word")
        if handle == "force":
            if re.match(r".*?_\d+$", s):
                i = re.match(r".*?_(\d+)$", s).groups()[0]
                s = s.strip("_" + i) + "_" + str(int(i) + 1)
            else:
                s += "_1"

    namespace[string] = s

    return s, namespace


def make_add_identifier(identifier, build_names):
    """
    Takes an existing used Python identifier and attatch a unique
    identifier with ADD_# ending.

    Used for add new information to an existing external object.
    build_names will be updated inside this functions as a set
    is mutable.

    Parameters
    ----------
    identifier: str
      Existing python identifier.

    build_names: set
      Set of the already used identifiers for external objects.

    Returns
    -------
    identifier: str
      A vaild python identifier based on the input indentifier
      and the existing ones.

    """
    identifier += "ADD_"
    number = 1
    # iterate until finding a non-used identifier
    while identifier + str(number) in build_names:
        number += 1

    # update identifier
    identifier += str(number)

    # update the build names
    build_names.add(identifier)

    return identifier


def get_return_elements(return_columns, namespace):
    """
    Takes a list of return elements formatted in vensim's format
    Varname[Sub1, SUb2]
    and returns first the model elements (in python safe language)
    that need to be computed and collected, and secondly the addresses
    that each element in the return columns list translates to

    Parameters
    ----------
    return_columns: list of strings

    namespace: dict

    Returns
    -------
    capture_elements
    return_addresses

    """
    capture_elements = list()
    return_addresses = dict()
    for col in return_columns:
        if col[0] == col[-1] and col[0] == '"':
            name = col
            address = None
        elif "[" in col:
            name, location = col.strip("]").split("[")
            address = tuple([loc.strip() for loc in location.split(",")])
        else:
            name = col
            address = None

        if name in namespace:
            py_name = namespace[name]
        else:
            if name in namespace.values():
                py_name = name
            else:
                raise KeyError(name + " not found as model element")

        if py_name not in capture_elements:
            capture_elements += [py_name]

        return_addresses[col] = (py_name, address)

    return list(capture_elements), return_addresses


def make_flat_df(df, return_addresses, flatten=False):
    """
    Takes a dataframe from the outputs of the integration processes,
    renames the columns as the given return_adresses and splits xarrays
    if needed.

    Parameters
    ----------
    df: Pandas.DataFrame
        Output from the integration.

    return_addresses: dict
        Keys will be column names of the resulting dataframe, and are what the
        user passed in as 'return_columns'. Values are a tuple:
        (py_name, {coords dictionary}) which tells us where to look for the
        value to put in that specific column.

    flatten: bool (optional)
            If True, once the output dataframe has been formatted will
            split the xarrays in new columns following vensim's naming
            to make a totally flat output. Default is False.

    Returns
    -------
    df: pandas.DataFrame
        Formatted dataframe.

    """
    cols_to_remove = set()
    rename_cols = {}
    for real_name, (pyname, address) in return_addresses.items():
        if address:
            cols_to_remove.add(pyname)
            # subset the value and add it to a new column
            xrval = [x.loc[address] for x in df[pyname].values]
            if xrval[0].size > 1:
                df[real_name] = xrval
            else:
                df[real_name] = [float(x) for x in xrval]
        else:
            # save the name to change it in the dataframe
            try:
                # some elements are returned as 0-d arrays, convert
                # them to float
                df[pyname] = [float(x) for x in df[pyname].values]
            except TypeError:
                pass
            rename_cols[pyname] = real_name

    df.rename(columns=rename_cols, inplace=True)

    if cols_to_remove:
        # remove the columns of the subset values
        df.drop(cols_to_remove, axis='columns', inplace=True)

    if flatten:
        # create a totally flat df (no xarray.DataArray)
        cols_to_remove.clear()
        for col in df.columns:
            if isinstance(df[col].values[0], xr.DataArray):
                # remove subscripts from name if given
                name = re.sub(r'\[.*\]', '', col)
                # split values in xarray.DataArray
                lval = [xrsplit(val) for val in df[col].values]
                for i, ar in enumerate(lval[0]):
                    vals = [float(v[i]) for v in lval]
                    subs = '[' + ','.join([str(ar.coords[dim].values)
                                           for dim in list(ar.coords)]) + ']'
                    df[name+subs] = vals
                    cols_to_remove.add(col)

        if cols_to_remove:
            # remove the columns of the subset values
            df.drop(cols_to_remove, axis='columns', inplace=True)

    return df


def compute_shape(coords, reshape_len=None, py_name=""):
    """
    Computes the 'shape' of a coords dictionary.
    Function used to rearange data in xarrays and
    to compute the number of rows/columns to be read in a file.

    Parameters
    ----------
    coords: dict
      Ordered dictionary of the dimension names as a keys with their values.

    reshape_len: int (optional)
      Number of dimensions of the output shape.
      The shape will ony compute the corresponent table
      dimensions to read from Excel, then, the dimensions
      with length one will be ignored at first.
      Lately, it will complete with 1 on the left of the shape
      if the reshape_len value is bigger than the length of shape.
      Will raise a ValueError if we try to reshape to a reshape_len
      smaller than the initial shape.

    py_name: str
      Name to print if an error is raised.

    Returns
    -------
    shape: list
      Shape of the ordered dictionary or of the desired table or vector.

    Note
    ----
    Dictionaries in Python >= 3.7 are ordered, which means that
    we could remove dims if there is a not backward compatible
    version of the library which only works in Python 3.7+. For now,
    the dimensions list is passed to make it work properly for all the users.

    """
    if not reshape_len:
        return [len(coord) for coord in coords.values()]

    # get the shape of the coordinates bigger than 1
    shape = [len(coord) for coord in coords.values() if len(coord) > 1]

    shape_len = len(shape)

    # return an error when the current shape is bigger than the requested one
    if shape_len > reshape_len:
        raise ValueError(
            py_name
            + "\n"
            + "The shape of the coords to read in a "
            + " external file must be at most "
            + "{} dimensional".format(reshape_len)
        )

    # complete with 1s on the left
    return [1] * (reshape_len - shape_len) + shape


def get_value_by_insensitive_key_or_value(key, dict):
    lower_key = key.lower()
    for real_key, real_value in dict.items():
        if real_key.lower() == lower_key:
            return dict[real_key]
        if real_value.lower() == lower_key:
            return real_value

    return None


def rearrange(data, dims, coords):
    """
    Returns a xarray.DataArray object with the given coords and dims

    Parameters
    ---------
    data: float or xarray.DataArray
        The input data to rearrange.

    dims: list
        Ordered list of the dimensions.

    coords: dict
        Dictionary of the dimension names as a keys with their values.

    Returns
    -------
    xarray.DataArray

    """
    # subset used coords in general coords will be the subscript_dict
    coords = {dim: coords[dim] for dim in dims}
    if isinstance(data, xr.DataArray):
        shape = tuple(compute_shape(coords))
        if data.shape == shape:
            # Allows switching dimensions names and transpositions
            return xr.DataArray(data=data.values, coords=coords, dims=dims)
        elif np.prod(shape) < np.prod(data.shape):
            # Allows subscripting a subrange
            return data.rename(
                {dim: new_dim for dim, new_dim in zip(data.dims, dims)}
            ).loc[coords]

        # The coordinates are expanded or transposed
        return xr.DataArray(0, coords, dims) + data

    elif data is not None:
        return xr.DataArray(data, coords, dims)

    return None


def round_(x):
    """
    Redefinition of round function to make it work with floats and xarrays
    """
    if isinstance(x, xr.DataArray):
        return x.round()

    return round(x)


def simplify_subscript_input(coords, subscript_dict, return_full, merge_subs):
    """
    Parameters
    ----------
    coords: dict
        Coordinates to write in the model file.

    subscript_dict: dict
        The subscript dictionary of the model file.

    return_full: bool
        If True the when coords == subscript_dict, '_subscript_dict'
        will be returned

    merge_subs: list of strings
        List of the final subscript range of the python array after
        merging with other objects

    Returns
    -------
    coords: str
        The equations to generate the coord dicttionary in the model file.

    """

    if coords == subscript_dict and return_full:
        # variable defined with all the subscripts
        return "_subscript_dict"

    coordsp = []
    for ndim, (dim, coord) in zip(merge_subs, coords.items()):
        # find dimensions can be retrieved from _subscript_dict
        if coord == subscript_dict[dim]:
            # use _subscript_dict
            coordsp.append(f"'{ndim}': _subscript_dict['{dim}']")
        else:
            # write whole dict
            coordsp.append(f"'{ndim}': {coord}")

    return "{" + ", ".join(coordsp) + "}"


def add_entries_underscore(*dictionaries):
    """
    Expands dictionaries adding new keys underscoring the white spaces
    in the old ones. As the dictionaries are mutable objects this functions
    will add the new entries to the already existing dictionaries with
    no need to return a new one.

    Parameters
    ----------
    *dictionaries: dict(s)
        The dictionary or dictionaries to add the entries with underscore.

    Return
    ------
    None

    """
    for dictionary in dictionaries:
        keys = list(dictionary)
        for name in keys:
            dictionary[re.sub(" ", "_", name)] = dictionary[name]
    return


def load_model_data(root_dir, model_name):

    """
    Used for models split in several files.
    Loads subscripts_dic, namespace and modules dictionaries

    Parameters
    ----------
    root_dir: str
        Path to the model file.

    model_name: str
        Name of the model without file type extension (e.g. "my_model").

    Returns
    -------
    namespace: dict
        Translation from original model element names (keys) to python safe
        function identifiers (values).

    subscripts: dict
        Dictionary describing the possible dimensions of the stock's
        subscripts.

    modules: dict
        Dictionary containing view (module) names as keys and a list of the
        corresponding variables as values.

    """

    with open(os.path.join(root_dir, "_subscripts_" + model_name + ".json")
              ) as subs:
        subscripts = json.load(subs)

    with open(os.path.join(root_dir, "_namespace_" + model_name + ".json")
              ) as names:
        namespace = json.load(names)

    # the _modules.json in the sketch_var folder shows to which module each
    # variable belongs
    with open(os.path.join(root_dir, "modules_" + model_name, "_modules.json")
              ) as mods:
        modules = json.load(mods)

    return namespace, subscripts, modules


def open_module(root_dir, model_name, module, submodule=None):
    """
    Used to load model modules from the main model file, when
    split_views=True in the read_vensim function.

    Parameters
    ----------
    root_dir: str
        Path to the model file.

    model_name: str
        Name of the model without file type extension (e.g. "my_model").

    module: str
        Name of the module folder or file to open.

    sub_module: str (optional)
        Name of the submodule to open.

    Returns
    -------
    str:
        Model file content.
    """
    if not submodule:
        rel_file_path = module + ".py"
    else:
        rel_file_path = os.path.join(module, submodule + ".py")

    return open(
        os.path.join(root_dir, "modules_" + model_name, rel_file_path)).read()


def clean_file_names(*args):
    """
    Removes special characters and makes clean file names

    Parameters
    ----------
    *args: tuple
        Any number of strings to to clean

    Returns
    -------
    clean: list
        List containing the clean strings
    """
    clean = []
    for name in args:
        clean.append(re.sub(
                            r"[\W]+", "", name.replace(" ", "_")
                            ).lstrip("0123456789")
                     )
    return clean


class ProgressBar:
    """
    Progress bar for integration
    """

    def __init__(self, maxval=None):

        self.maxval = maxval
        if self.maxval is None:
            return

        self.counter = 0

        self.bar = progressbar.ProgressBar(
            maxval=self.maxval,
            widgets=[
                progressbar.ETA(),
                " ",
                progressbar.Bar("#", "[", "]", "-"),
                progressbar.Percentage(),
            ],
        )

        self.bar.start()

    def update(self):
        """Update progress bar"""
        try:
            self.counter += 1
            self.bar.update(self.counter)
        except AttributeError:
            # Error if bar is not imported
            pass

    def finish(self):
        """Finish progress bar"""
        try:
            self.bar.finish()
        except AttributeError:
            # Error if bar is not imported
            pass



"""
Deals with accessing the components of the xmile file, and
formatting them for the builder

James Houghton <james.p.houghton@gmail.com>
Alexey Prey Mulyukin <alexprey@yandex.ru> from sdCloud.io development team.

"""
import re
import os.path

from .SMILE2Py import SMILEParser
from lxml import etree
from .. import builder, utils

import numpy as np


def translate_xmile(xmile_file):
    """ Translate an xmile model file into a python class.
    Functionality is currently limited.

    """

    # process xml file
    xml_parser = etree.XMLParser(encoding="utf-8", recover=True)
    root = etree.parse(xmile_file, parser=xml_parser).getroot()
    NS = root.nsmap[None]  # namespace of the xmile document

    def get_xpath_text(node, path, ns=None, default=''):
        """ Safe access of occassionally missing elements """
        # defined here to take advantage of NS in default
        if ns is None:
            ns = {'ns': NS}
        try:
            return node.xpath(path, namespaces=ns)[0].text
        except:
            return default

    def get_xpath_attrib(node, path, attrib, ns=None, default=None):
        """ Safe access of occassionally missing elements """
        # defined here to take advantage of NS in default
        if ns is None:
            ns = {'ns': NS}
        try:
            return node.xpath(path, namespaces=ns)[0].attrib[attrib]
        except:
            return default

    def is_constant_expression(py_expr):
        try:
            val = float(py_expr)
            return True
        except ValueError:
            return False

    def parse_lookup_xml_node(node):
        ys_node = node.xpath('ns:ypts', namespaces={'ns': NS})[0]
        ys = np.fromstring(ys_node.text, dtype=float, sep=ys_node.attrib['sep'] if 'sep' in ys_node.attrib else ',')
        xscale_node = node.xpath('ns:xscale', namespaces={'ns': NS})
        if len(xscale_node) > 0:
            xmin = xscale_node[0].attrib['min']
            xmax = xscale_node[0].attrib['max']
            xs = np.linspace(float(xmin), float(xmax), len(ys))
        else:
            xs_node = node.xpath('ns:xpts', namespaces={'ns': NS})[0]
            xs = np.fromstring(xs_node.text, dtype=float, sep=xs_node.attrib['sep'] if 'sep' in xs_node.attrib else ',')

        type = node.attrib['type'] if 'type' in node.attrib else 'continuous'

        functions_map = {
            "continuous": {
                "name": "lookup",
                "module": "functions"
            },
            'extrapolation': {
                "name": "lookup_extrapolation",
                "module": "functions"
            },
            'discrete': {
                "name": "lookup_discrete",
                "module": "functions"
            }
        }
        lookup_function = functions_map[type] if type in functions_map else functions_map['continuous']

        return {
            'name': node.attrib['name'] if 'name' in node.attrib else '',
            'xs': xs,
            'ys': ys,
            'type': type,
            'function': lookup_function
        }

    # build model namespace
    namespace = {
        'TIME': 'time',
        'Time': 'time',
        'time': 'time'
    }  # namespace of the python model
    names_xpath = '//ns:model/ns:variables/ns:aux|' \
                  '//ns:model/ns:variables/ns:flow|' \
                  '//ns:model/ns:variables/ns:stock|' \
                  '//ns:model/ns:variables/ns:gf'

    for node in root.xpath(names_xpath, namespaces={'ns': NS}):
        name = node.attrib['name']
        _, namespace = utils.make_python_identifier(name, namespace)

    model_elements = []
    smile_parser = SMILEParser(namespace)

    # add aux and flow elements
    flaux_xpath = '//ns:model/ns:variables/ns:aux|//ns:model/ns:variables/ns:flow'
    for node in root.xpath(flaux_xpath, namespaces={'ns': NS}):
        name = node.attrib['name']
        units = get_xpath_text(node, 'ns:units')
        lims = (get_xpath_attrib(node, 'ns:range', 'min'), get_xpath_attrib(node, 'ns:range', 'max'))
        lims = str(tuple(float(x) if x is not None else x for x in lims))
        doc = get_xpath_text(node, 'ns:doc')
        py_name = namespace[name]
        eqn = get_xpath_text(node, 'ns:eqn')

        # Replace new lines with space, and replace 2 or more spaces with single space
        # Then ensure there is no space at start or end of equation
        eqn = (re.sub(r"(\s{2,})", " ", eqn.replace("\n", ' '))
                 .strip()
        )

        element = {
            'kind': 'component',
            'real_name': name,
            'unit': units,
            'doc': doc,
            'eqn': eqn,
            'lims': lims,
            'py_name': py_name,
            'subs': [],  # Todo later
            'arguments': '',
        }

        tranlation, new_structure = smile_parser.parse(eqn, element)
        element.update(tranlation)
        if is_constant_expression(element['py_expr']):
            element['kind'] = 'constant'

        model_elements += new_structure

        gf_node = node.xpath("ns:gf", namespaces={'ns': NS})
        if len(gf_node) > 0:
            gf_data = parse_lookup_xml_node(gf_node[0])
            xs = '[' + ','.join("%10.3f" % x for x in gf_data['xs']) + ']'
            ys = '[' + ','.join("%10.3f" % x for x in gf_data['ys']) + ']'
            py_expr =\
                builder.build_function_call(gf_data['function'],
                                            [element['py_expr'], xs, ys])\
                + ' if x is None else '\
                + builder.build_function_call(gf_data['function'],
                                              ['x', xs, ys])
            element.update({
                'kind': 'lookup',
                # This lookup declared as inline, so we should implement inline mode for flow and aux
                'arguments': "x = None",
                'py_expr': py_expr,
            })

        model_elements.append(element)

    # add gf elements
    gf_xpath = '//ns:model/ns:variables/ns:gf'
    for node in root.xpath(gf_xpath, namespaces={'ns': NS}):
        name = node.attrib['name']
        py_name = namespace[name]

        units = get_xpath_text(node, 'ns:units')
        doc = get_xpath_text(node, 'ns:doc')

        gf_data = parse_lookup_xml_node(node)
        xs = '[' + ','.join("%10.3f" % x for x in gf_data['xs']) + ']'
        ys = '[' + ','.join("%10.3f" % x for x in gf_data['ys']) + ']'
        py_expr = builder.build_function_call(gf_data['function'],
                                              ['x', xs, ys])
        element = {
            'kind': 'lookup',
            'real_name': name,
            'unit': units,
            'lims': None,
            'doc': doc,
            'eqn': '',
            'py_name': py_name,
            'py_expr': py_expr,
            'arguments': 'x',
            'subs': [],  # Todo later
        }
        model_elements.append(element)

    # add stock elements
    stock_xpath = '//ns:model/ns:variables/ns:stock'
    for node in root.xpath(stock_xpath, namespaces={'ns': NS}):
        name = node.attrib['name']
        units = get_xpath_text(node, 'ns:units')
        lims = (get_xpath_attrib(node, 'ns:range', 'min'), get_xpath_attrib(node, 'ns:range', 'max'))
        lims = str(tuple(float(x) if x is not None else x for x in lims))
        doc = get_xpath_text(node, 'ns:doc')
        py_name = namespace[name]

        # Extract input and output flows equations
        inflows = [n.text for n in node.xpath('ns:inflow', namespaces={'ns': NS})]
        outflows = [n.text for n in node.xpath('ns:outflow', namespaces={'ns': NS})]

        eqn = ' + '.join(inflows) if inflows else ''
        eqn += (' - ' + ' - '.join(outflows)) if outflows else ''

        element = {
            'kind': 'component' if inflows or outflows else 'constant',
            'real_name': name,
            'unit': units,
            'doc': doc,
            'eqn': eqn,
            'lims': lims,
            'py_name': py_name,
            'subs': [],  # Todo later
            'arguments': ''
        }

        # Parse each flow equations
        py_inflows = []
        for inputFlow in inflows:
            translation, new_structure = smile_parser.parse(inputFlow, element)
            py_inflows.append(translation['py_expr'])
            model_elements += new_structure

        # Parse each flow equations
        py_outflows = []
        for outputFlow in outflows:
            translation, new_structure = smile_parser.parse(outputFlow, element)
            py_outflows.append(translation['py_expr'])
            model_elements += new_structure

        py_ddt = ' + '.join(py_inflows) if py_inflows else ''
        py_ddt += (' - ' + ' - '.join(py_outflows)) if py_outflows else ''

        # Read the initial value equation for stock element
        initial_value_eqn = get_xpath_text(node, 'ns:eqn')
        translation, new_structure = smile_parser.parse(initial_value_eqn, element)
        py_initial_value = translation['py_expr']
        model_elements += new_structure

        py_expr, new_structure = builder.add_stock(identifier=py_name,
                                                   subs=[],  # Todo later
                                                   merge_subs=[],
                                                   expression=py_ddt,
                                                   initial_condition=py_initial_value
                                                   )
        element['py_expr'] = py_expr
        model_elements.append(element)
        model_elements += new_structure

    # remove timestamp pieces so as not to double-count
    model_elements_parsed = []
    for element in model_elements:
        if element['real_name'].lower() not in ['initial time', 'final time', 'time step', 'saveper']:
            model_elements_parsed.append(element)
    model_elements = model_elements_parsed

    # Add timeseries information

    # Read the start time of simulation
    sim_spec_node = root.xpath('//ns:sim_specs', namespaces={'ns': NS});
    time_units = sim_spec_node[0].attrib['time_units'] if (len(sim_spec_node) > 0 and 'time_units' in sim_spec_node[0].attrib) else ""

    tstart = root.xpath('//ns:sim_specs/ns:start', namespaces={'ns': NS})[0].text
    element = {
        'kind': 'constant',
        'real_name': 'INITIAL TIME',
        'unit': time_units,
        'lims': None,
        'doc': 'The initial time for the simulation.',
        'eqn': tstart,
        'py_name': 'initial_time',
        'subs': None,
        'arguments': '',
    }
    translation, new_structure = smile_parser.parse(tstart, element)
    element.update(translation)
    model_elements.append(element)
    model_elements += new_structure

    # Read the final time of simulation
    tstop = root.xpath('//ns:sim_specs/ns:stop', namespaces={'ns': NS})[0].text
    element = {
        'kind': 'constant',
        'real_name': 'FINAL TIME',
        'unit': time_units,
        'lims': None,
        'doc': 'The final time for the simulation.',
        'eqn': tstart,
        'py_name': 'final_time',
        'subs': None,
        'arguments': '',
    }

    translation, new_structure = smile_parser.parse(tstop, element)
    element.update(translation)
    model_elements.append(element)
    model_elements += new_structure

    # Read the time step of simulation
    dt_node = root.xpath('//ns:sim_specs/ns:dt', namespaces={'ns': NS})

    # Use default value for time step if `dt` is not specified in model
    dt_eqn = "1.0"
    if len(dt_node) > 0:
        dt_node = dt_node[0]
        dt_eqn = dt_node.text
        # If reciprocal mode are defined for `dt`, we should inverse value
        if ("reciprocal" in dt_node.attrib and dt_node.attrib["reciprocal"].lower() == "true"):
            dt_eqn = "1/" + dt_eqn

    element = {
        'kind': 'constant',
        'real_name': 'TIME STEP',
        'unit': time_units,
        'lims': None,
        'doc': 'The time step for the simulation.',
        'eqn': dt_eqn,
        'py_name': 'time_step',
        'subs': None,
        'arguments': '',
    }
    translation, new_structure = smile_parser.parse(dt_eqn, element)
    element.update(translation)
    model_elements.append(element)
    model_elements += new_structure

    # Add the SAVEPER attribute to the model
    model_elements.append({
        'kind': 'constant',
        'real_name': 'SAVEPER',
        'unit': time_units,
        'lims': None,
        'doc': 'The time step for the simulation.',
        'eqn': dt_eqn,
        'py_name': 'saveper',
        'py_expr': 'time_step()',
        'subs': None,
        'arguments': '',
    })

    file_name, file_extension = os.path.splitext(xmile_file)
    outfile_name = file_name + '.py'

    builder.build(elements=model_elements,
                  subscript_dict={},
                  namespace=namespace,
                  outfile_name=outfile_name)

    return outfile_name



from setuptools import setup, find_packages

exec(open('pysd/_version.py').read())
print(__version__)

test_pckgs = open('dev-requirements.txt').read().strip().split('\n')

setup(
    name='pysd',
    version=__version__,
    python_requires='>=3.7',
    author='James Houghton',
    author_email='james.p.houghton@gmail.com',
    packages=find_packages(exclude=['docs', 'tests', 'dist', 'build']),
    url='https://github.com/JamesPHoughton/pysd',
    license='LICENSE.txt',
    description='System Dynamics Modeling in Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['System Dynamics', 'Vensim', 'XMILE'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Intended Audience :: Science/Research',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=open('requirements.txt').read().strip().split('\n'),
    tests_require=test_pckgs,
    extras_require={"test": test_pckgs},
    package_data={
        'py_backend': [
            'xmile/smile.grammar'
        ]
    },
    include_package_data=True
)



from .pysd import read_vensim, read_xmile, load
from .py_backend import functions, utils, external
from .py_backend.decorators import cache, subs
from ._version import __version__
from .py_backend.vensim.table2py import read_tabular

